/*
 * $Id: oftpd.c,v 1.12 2000/03/29 01:20:31 shane Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <pwd.h>
#include <syslog.h>
#include <pthread.h>
#include "oftpd.h"
#include "ftp_listener.h"
#include "error.h"

/* put our executable name here where everybody can see it */
static const char *exe_name = "oftpd";

static void daemonize();

int main(int argc, char *argv[])
{
    struct passwd *user_info;
    error_t err;

    ftp_listener_t ftp_listener;

    /* grab our executable name */
    if (argc > 0) {
        exe_name = argv[0];
    }

    /* verify we're running as root */
    if (geteuid() != 0) {
        fprintf(stderr, "%s: program needs root permission to run\n", exe_name);
        exit(1);
    }

    /* check our command-line arguments */
    if (argc != 3) {
        fprintf(stderr, "%s: syntax error\n", exe_name);
	fprintf(stderr, " Syntax: %s user_name root_directory\n", exe_name);
        exit(1);
    }
    user_info = getpwnam(argv[1]);
    if (user_info == NULL) {
        fprintf(stderr, "%s: invalid user name\n", exe_name);
        exit(1);
    }

    /* become a daemon */
    daemonize();

    /* change to root directory */
    if (chroot(argv[2]) != 0) {
        syslog(LOG_ERR, "error with root directory; %s\n", exe_name, 
	  strerror(errno));
        exit(1);
    }
    if (chdir("/") != 0) {
        syslog(LOG_ERR, "error changing directory; %s\n", strerror(errno));
	exit(1);
    }

    /* create our main listener */
    if (!ftp_listener_init(&ftp_listener, 
                           FTP_ADDRESS, 
			   FTP_PORT, 
			   MAX_CLIENTS, 
			   INACTIVITY_TIMEOUT, 
			   &err)) 
    {
        syslog(LOG_ERR, "error initializing FTP listener; %s",
	  error_get_desc(&err));
        exit(1);
    }

    /* set user to be as inoffensive as possible */
    setgid(user_info->pw_gid);
    setuid(user_info->pw_uid);

    /* start our listener */
    if (ftp_listener_start(&ftp_listener) == 0) {
        syslog(LOG_ERR, "error starting FTP service");
        exit(1);
    }

    /* done */
    pthread_exit(NULL);
    return 0;
}

static void daemonize()
{
    int fork_ret;
    int max_fd;
    int null_fd;
    int fd;

    null_fd = open("/dev/null", O_RDWR);
    if (null_fd == -1) {
        fprintf(stderr, "%s: error opening null output device; %s\n", exe_name, 
	  strerror(errno));
	exit(1);
    }

    max_fd = sysconf(_SC_OPEN_MAX);
    if (max_fd == -1) {
        fprintf(stderr, "%s: error getting maximum open file; %s\n", exe_name, 
	  strerror(errno));
	exit(1);
    }


    fork_ret = fork();
    if (fork_ret == -1) {
        fprintf(stderr, "%s: error forking; %s\n", exe_name, strerror(errno));
	exit(1);
    }
    if (fork_ret != 0) {
        exit(0);
    }
    if (setsid() == -1) {
        fprintf(stderr, "%s: error creating process group; %s\n", exe_name, 
	  strerror(errno));
	exit(1);
    }
    fork_ret = fork();
    if (fork_ret == -1) {
        fprintf(stderr, "%s: error forking; %s\n", exe_name, strerror(errno));
	exit(1);
    }
    if (fork_ret != 0) {
        exit(0);
    }
    if (dup2(null_fd, 0) == -1) {
        syslog(LOG_ERR, "error setting input to null; %s", 
	  strerror(errno));
	exit(1);
    }
    if (dup2(null_fd, 1) == -1) {
        syslog(LOG_ERR, "error setting output to null; %s", 
	  strerror(errno));
	exit(1);
    }
    if (dup2(null_fd, 2) == -1) {
        syslog(LOG_ERR, "error setting error output to null; %s", 
	  strerror(errno));
	exit(1);
    }
    for (fd=3; fd<max_fd; fd++) {
        close(fd);
    }

    signal(SIGPIPE, SIG_IGN);         /* avoid SIGPIPE on socket activity */
    openlog(NULL, LOG_NDELAY, LOG_DAEMON);
}

